Imports System.Math

Module FftDouble
    Public Const FFT_FORWARD As Integer = 1
    Public Const FFT_REVERSE As Integer = -1

    Public Sub Four1(ByVal Data() As Double, ByVal Isign As Integer)
        Dim n As Integer = Data.Length

        Dim h As Integer = 0
        For i As Integer = 0 To n - 2 Step 2
            If h > i Then
                Dim temp As Double
                temp = Data(h) : Data(h) = Data(i) : Data(i) = temp
                temp = Data(h + 1) : Data(h + 1) = Data(i + 1) : Data(i + 1) = temp
            End If

            Dim m As Integer = n >> 1
            Do While (m >= 2 AndAlso h > m - 1)
                h -= m
                m >>= 1
            Loop
            h += m
        Next

        Dim mmax As Integer = 2
        Do While n > mmax
            Dim istep As Integer = mmax * 2
            Dim theta As Double = Isign * (PI * 2.0# / mmax)
            Dim sinth As Double = Sin(0.5# * theta)
            Dim wpr As Double = -2.0# * sinth * sinth
            Dim wpi As Double = Sin(theta)
            Dim wr As Double = 1.0#
            Dim wi As Double = 0.0#

            For m As Integer = 1 To mmax - 1 Step 2
                For i As Integer = m - 1 To n - 1 Step istep
                    Dim j As Integer = i + mmax
                    Dim tempr As Double = wr * Data(j) - wi * Data(j + 1)
                    Dim tempi As Double = wr * Data(j + 1) + wi * Data(j)
                    Data(j) = Data(i) - tempr
                    Data(j + 1) = Data(i + 1) - tempi
                    Data(i) += tempr
                    Data(i + 1) += tempi
                Next

                Dim wrt As Double = wr
                wr = wrt * wpr - wi * wpi + wr
                wi = wi * wpr + wrt * wpi + wi
            Next

            mmax = istep
        Loop
    End Sub

    Public Sub TwoFft(ByVal Data1() As Double, ByVal Data2() As Double, _
                      ByRef Fft1() As Double, ByRef Fft2() As Double)
        Dim n As Integer = Data1.Length
        Dim nn As Integer = 2 * n
        Dim nn1 As Integer = nn + 1

        ReDim Fft1(nn - 1), Fft2(nn - 1)

        Dim jj As Integer = 1
        For j As Integer = 0 To n - 1
            Fft1(jj - 1) = Data1(j)
            Fft1(jj) = Data2(j)
            jj += 2
        Next

        Four1(Fft1, FFT_FORWARD)
        Fft2(0) = Fft1(1)
        Fft1(1) = 0.0#
        Fft2(1) = 0.0#

        For j As Integer = 2 To n Step 2
            Dim realp As Double = 0.5# * (Fft1(j) + Fft1(nn - j))
            Dim realm As Double = 0.5# * (Fft1(j) - Fft1(nn - j))
            Dim aip As Double = 0.5# * (Fft1(j + 1) + Fft1(nn1 - j))
            Dim aim As Double = 0.5# * (Fft1(j + 1) - Fft1(nn1 - j))

            Fft1(j) = realp
            Fft1(j + 1) = aim

            Fft1(nn - j) = realp
            Fft1(nn1 - j) = -aim

            Fft2(j) = aip
            Fft2(j + 1) = -realm

            Fft2(nn - j) = aip
            Fft2(nn1 - j) = realm
        Next
    End Sub

    Public Sub RealFt(ByVal Data() As Double, ByVal Isign As Integer)
        Dim n As Integer = Data.Length

        Dim c1 As Double = 0.5#
        Dim c2 As Double

        Dim theta As Double = PI / (n >> 1)
        If Isign = FFT_FORWARD Then
            c2 = -0.5#
            Four1(Data, FFT_FORWARD)
        Else
            c2 = 0.5#
            theta = -theta
        End If

        Dim sinth As Double = Sin(0.5# * theta)
        Dim wpr As Double = -2.0# * sinth * sinth
        Dim wpi As Double = Sin(theta)
        Dim wr As Double = 1.0# + wpr
        Dim wi As Double = wpi

        For i As Integer = 1 To (n >> 2) - 1
            Dim i1 As Integer = i + i
            Dim i2 As Integer = i1 + 1
            Dim i3 As Integer = n - i1
            Dim i4 As Integer = i3 + 1

            Dim h1r As Double = c1 * (Data(i1) + Data(i3))
            Dim h1i As Double = c1 * (Data(i2) - Data(i4))
            Dim h2r As Double = -c2 * (Data(i2) + Data(i4))
            Dim h2i As Double = c2 * (Data(i1) - Data(i3))

            Data(i1) = h1r + wr * h2r - wi * h2i
            Data(i2) = h1i + wr * h2i + wi * h2r
            Data(i3) = h1r - wr * h2r + wi * h2i
            Data(i4) = -h1i + wr * h2i + wi * h2r

            Dim wrt As Double = wr
            wr = wrt * wpr - wi * wpi + wr
            wi = wi * wpr + wrt * wpi + wi
        Next

        If Isign = FFT_FORWARD Then
            Dim data0 As Double = Data(0)
            Data(0) = data0 + Data(1)
            Data(1) = data0 - Data(1)
        Else
            Dim data0 As Double = Data(0)
            Data(0) = c1 * (data0 + Data(1))
            Data(1) = c1 * (data0 - Data(1))
            Four1(Data, FFT_REVERSE)
        End If
    End Sub
End Module
